'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _command = require('@oclif/command');

var _liskElements = require('lisk-elements');

var _liskElements2 = _interopRequireDefault(_liskElements);

var _base = require('../../../base');

var _base2 = _interopRequireDefault(_base);

var _flags = require('../../../utils/flags');

var _flags2 = _interopRequireDefault(_flags);

var _input = require('../../../utils/input');

var _input2 = _interopRequireDefault(_input);

var _helpers = require('../../../utils/helpers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
 * LiskHQ/lisk-commander
 * Copyright © 2017–2018 Lisk Foundation
 *
 * See the LICENSE file at the top-level directory of this distribution
 * for licensing information.
 *
 * Unless otherwise agreed in a custom licensing agreement with the Lisk Foundation,
 * no part of this software, including this file, may be copied, modified,
 * propagated, or distributed except according to the terms contained in the
 * LICENSE file.
 *
 * Removal or modification of this copyright notice is prohibited.
 *
 */
const processInputs = (lifetime, minimum, keysgroup) => ({
	passphrase,
	secondPassphrase
}) => _liskElements2.default.transaction.registerMultisignature({
	passphrase,
	secondPassphrase,
	keysgroup,
	lifetime,
	minimum
});

class MultisignatureCommand extends _base2.default {
	async run() {
		const {
			args: { lifetime, minimum, keysgroup },
			flags: {
				passphrase: passphraseSource,
				'second-passphrase': secondPassphraseSource,
				'no-signature': noSignature
			}
		} = this.parse(MultisignatureCommand);

		_liskElements2.default.transaction.utils.validatePublicKeys(keysgroup);

		(0, _helpers.validateLifetime)(lifetime);
		(0, _helpers.validateMinimum)(minimum);

		const transactionLifetime = parseInt(lifetime, 10);
		const transactionMinimumConfirmations = parseInt(minimum, 10);
		const processFunction = processInputs(transactionLifetime, transactionMinimumConfirmations, keysgroup);

		if (noSignature) {
			const result = processFunction({
				passphrase: null,
				secondPassphrase: null
			});
			return this.print(result);
		}

		const inputs = await (0, _input2.default)({
			passphrase: {
				source: passphraseSource,
				repeatPrompt: true
			},
			secondPassphrase: !secondPassphraseSource ? null : {
				source: secondPassphraseSource,
				repeatPrompt: true
			}
		});
		const result = processFunction(inputs);
		return this.print(result);
	}
}

exports.default = MultisignatureCommand;
MultisignatureCommand.args = [{
	name: 'lifetime',
	required: true,
	description: 'Number of hours the transaction should remain in the transaction pool before becoming invalid.'
}, {
	name: 'minimum',
	required: true,
	description: 'Minimum number of signatures required for a transaction from the account to be valid.'
}, {
	name: 'keysgroup',
	required: true,
	description: 'Public keys to verify signatures against for the multisignature group.',
	parse: input => input.split(',')
}];

MultisignatureCommand.flags = Object.assign({}, _base2.default.flags, {
	passphrase: _command.flags.string(_flags2.default.passphrase),
	'second-passphrase': _command.flags.string(_flags2.default.secondPassphrase),
	'no-signature': _command.flags.boolean(_flags2.default.noSignature)
});

MultisignatureCommand.description = `
Creates a transaction which will register the account as a multisignature account if broadcast to the network, using the following arguments:
	1. Number of hours the transaction should remain in the transaction pool before becoming invalid.
	2. Minimum number of signatures required for a transaction from the account to be valid.
	3. Public keys to verify signatures against for the multisignature group.
`;

MultisignatureCommand.examples = ['transaction:create:multisignature 24 2 215b667a32a5cd51a94c9c2046c11fffb08c65748febec099451e3b164452bca,922fbfdd596fa78269bbcadc67ec2a1cc15fc929a19c462169568d7a3df1a1aa'];